/*
 [The "BSD licence"]
 Copyright (c) 2005 Jean Bovet, http://www.cs.usfca.edu/~jbovet/
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
														   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
														   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "EthernalDocument.h"
#import "Ethernal.h"

#import "EEditorView.h"
#import "Sniffer.h"
#import "PriviledgedSocketCreator.h"
#import "EthernetPacket.h"
#import "ResultArrayController.h"

#import "GetEthernetAddrSample.h"

@interface EthernalDocument (PrivateMethods)
- (void)updateInfo;
- (void)updateStats;
@end

@implementation EthernalDocument

- (id)init
{
	if(self = [super init]) {
		mSniffer = [[Sniffer alloc] init];
		[mSniffer setDelegate:self];
		
		mUpdateTimer = NULL;
		mPacketBufferLock = [[NSLock alloc] init];
		mPacketBuffer = [[NSMutableArray alloc] init];
		
		mRunning = NO;
		mData = NULL;
	}
	return self;
}

- (void)dealloc
{
	[mPacketBuffer release];
	[mPacketBufferLock release];
	[mSniffer release];
	[super dealloc];
}

- (id)data
{
	NSMutableDictionary *dic = [NSMutableDictionary dictionary];
	[dic setObject:[NSNumber numberWithInt:1] forKey:@"version"];

	NSMutableArray *array = [NSMutableArray array];
	NSEnumerator *enumerator = [[mResultController content] objectEnumerator];
	EthernetPacket *packet;
	while(packet = [enumerator nextObject]) {
		[array addObject:[packet xmlData]];
	}
	[dic setObject:array forKey:@"packets"];
	
	return dic;
}

- (void)loadData:(id)data
{
	[mResultController removeObjects:[mResultController content]];
	
	int index = 0;
	NSEnumerator *enumerator = [[data objectForKey:@"packets"] objectEnumerator];
	id xmlData;
	while(xmlData = [enumerator nextObject]) {
		EthernetPacket *packet = [EthernetPacket packetWithXMLData:xmlData];
		[packet setAbsoluteIndex:++index];
		[mResultController addObject:packet];
	}	
}

- (BOOL)writeToFile:(NSString *)fileName ofType:(NSString *)docType
{
	return [[self data] writeToFile:fileName atomically:YES];
}

- (BOOL)readFromFile:(NSString *)fileName ofType:(NSString *)docType
{
	mData = [NSMutableDictionary dictionaryWithContentsOfFile:fileName];
	return mData != nil;
}

#pragma mark -

- (void)removeWindowController:(NSWindowController *)windowController
{
	[mUpdateTimer invalidate];
	[mSniffer stop];
	[super removeWindowController:windowController];
}

- (void)windowControllerDidLoadNib:(NSWindowController *)windowController
{
	[mInfoController addObject:[NSDictionary dictionaryWithObject:[NSNumber numberWithBool:mRunning] forKey:@"running"]];
	
	EnetData data[10];
	UInt32 entryCount = 10;
	int err = GetEthernetAddressInfo((EnetData*)&data, &entryCount);
	if(err == noErr) {
		int i;
		for(i=0; i<entryCount; i++) {
			NSMutableString *title = [NSMutableString string];
			[title appendFormat:@"%s", data[i].bsdName];
			//	if(data[i].isBuiltIn)
			//		[title appendString:@" Built-in"];
			[mSlotCombo insertItemWithObjectValue:title atIndex:0];
			if(i == entryCount-1)
				[mSlotCombo setStringValue:title];
		}
	}
	
	NSString *defaultSlot = [[NSUserDefaults standardUserDefaults] objectForKey:@"slot"];
	if(defaultSlot)
		[mSlotCombo setStringValue:defaultSlot];
	
	[self loadData:mData];
	[self updateStats];
	[self updateInfo];
	
	mUpdateTimer = [NSTimer scheduledTimerWithTimeInterval:0.25 target:self selector:@selector(updatePackets) userInfo:nil repeats:YES];
	[[NSRunLoop currentRunLoop] addTimer:mUpdateTimer forMode:NSEventTrackingRunLoopMode];
}

- (NSString *)windowNibName
{
    return @"Document";
}

#pragma mark -

- (void)updateInfo
{
	NSMutableString *info = [NSMutableString string];
	
	int total = [[mResultController content] count];
	int arranged = [[mResultController arrangedObjects] count];
	if(total == arranged)
		[info appendFormat:@"%d packets", total];
	else
		[info appendFormat:@"%d of %d packets", arranged, total];
		
	[mInfoTextField setStringValue:info];	
}

- (void)updateStats
{
	NSMutableString *stat = [NSMutableString stringWithString:@"System: "];
	
	unsigned received = [mSniffer packetsReceived];
	unsigned dropped = [mSniffer packetsDropped];
	if(received > 1)
		[stat appendFormat:@"%d packets received", received];
	else
		[stat appendFormat:@"%d packet received", received];
	
	if(dropped > 0) {
		if(dropped == 1)
			[stat appendFormat:@", %d packet dropped", dropped];
		else
			[stat appendFormat:@", %d packets dropped", dropped];
	}
	
	[mStatsTextField setStringValue:stat];
}

- (BOOL)prepareSniffer
{
	int socket = [[Ethernal shared] getSocketForSlot:[mSlotCombo stringValue]];
	if(socket == -1)
		return NO;
	else {
		[mSniffer setBPFSocket:socket];	
		return YES;
	}	
}

- (void)applyCurrentSelection
{
	if([[mResultController selectedObjects] count] > 0) {
		[mEditorView setContent:[[mResultController selectedObjects] objectAtIndex:0]];		
		[mEditorView setHighlightString:[mSearchField stringValue]];
	} else {
		[mEditorView setContent:NULL];
	}
}

#pragma mark -

- (IBAction)changeSlot:(id)sender
{
	[[NSUserDefaults standardUserDefaults] setObject:[mSlotCombo stringValue] forKey:@"slot"];
}

- (IBAction)startStopAction:(id)sender
{
	if(mRunning) {
		mRunning = NO;
		[mSniffer quiet:YES];
		[mStartStopButton setTitle:@"Start"];
	} else {
		if([self prepareSniffer]) {
			mRunning = YES;			
			[mSniffer start];
			[mSniffer quiet:NO];
			[mStartStopButton setTitle:@"Stop"];
		}
	}
	[mInfoController addObject:[NSDictionary dictionaryWithObject:[NSNumber numberWithBool:mRunning] forKey:@"running"]];
}

- (IBAction)zoomAction:(id)sender
{
	[mEditorView setCharactersWidth:[sender intValue]];
	[self applyCurrentSelection];
}

- (IBAction)search:(id)sender
{
	[mResultController setSearchString:[sender stringValue]];
	[mResultController rearrangeObjects];
	[mEditorView setHighlightString:[sender stringValue]];
	[self updateInfo];
}

- (IBAction)changePacketsRepresentation:(id)sender
{
	[mResultController rearrangeObjects];
}

#pragma mark -

- (NSWindow*)window
{
	return [[[self windowControllers] objectAtIndex:0] window];
}

- (void)performExportToFile:(NSString*)file
{
	BOOL exportAll = [[NSUserDefaults standardUserDefaults] objectForKey:@"exportMode"] == 0;
	char delimiter;
	
	if([[NSUserDefaults standardUserDefaults] objectForKey:@"exportDelimiter"] == 0)
		delimiter = ',';
	else
		delimiter = '\t';
	
	BOOL exportHexaData = [[[NSUserDefaults standardUserDefaults] objectForKey:@"exportHexaData"] boolValue];
	BOOL exportASCIIData = [[[NSUserDefaults standardUserDefaults] objectForKey:@"exportASCIIData"] boolValue];

	NSMutableString *text = [NSMutableString string];
	
	NSEnumerator *enumerator = [(exportAll?[mResultController content]:[mResultController selectedObjects]) objectEnumerator];
	EthernetPacket *packet;
	while(packet = [enumerator nextObject]) {
		[text appendFormat:@"%d", [packet absoluteIndex]];
		[text appendFormat:@"%c", delimiter];

		[text appendString:[packet time]];
		[text appendFormat:@"%c", delimiter];

		[text appendString:[packet source]];
		[text appendFormat:@"%c", delimiter];

		[text appendString:[packet destination]];
		[text appendFormat:@"%c", delimiter];

		[text appendString:[packet sourceIP]];
		[text appendFormat:@"%c", delimiter];
		
		[text appendString:[packet destinationIP]];
		[text appendFormat:@"%c", delimiter];
		
		[text appendString:[packet protocol]];
		[text appendFormat:@"%c", delimiter];
		
		[text appendFormat:@"%d", [packet size]];

		if(exportHexaData) {
			[text appendFormat:@"%c", delimiter];
			[text appendString:[packet rawHexaContent]];
		}

		if(exportASCIIData) {
			[text appendFormat:@"%c", delimiter];
			[text appendString:[packet rawAsciiContentWithoutControlCharacters]];
		}		
		
		[text appendString:@"\r"];
	}
	
	[text writeToFile:file atomically:YES];
}

- (void)exportChooseFile
{
	NSSavePanel *panel = [NSSavePanel savePanel];
	[panel setRequiredFileType:@"txt"];
	[panel beginSheetForDirectory:NULL
							 file:NULL
				   modalForWindow:[self window] 
					modalDelegate:self
				   didEndSelector:@selector(chooseExportPacketsPanelDidEnd:returnCode:contextInfo:) 
					  contextInfo:nil];
	
}

- (void)chooseExportPacketsPanelDidEnd:(NSOpenPanel *)panel returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
	if(returnCode != NSOKButton)
		return;
	
	[self performExportToFile:[panel filename]];
}

- (IBAction)export:(id)sender
{	
	[NSApp beginSheet:mExportPanel modalForWindow:[self window] modalDelegate:self didEndSelector:NULL contextInfo:NULL];
}

- (IBAction)cancelExport:(id)sender
{
	[NSApp endSheet:mExportPanel];
	[mExportPanel orderOut:self];
}

- (IBAction)okExport:(id)sender
{
	[NSApp endSheet:mExportPanel];
	[mExportPanel orderOut:self];
	[self performSelector:@selector(exportChooseFile) withObject:NULL afterDelay:0];
}

#pragma mark -

// Update timer

- (void)updatePackets
{
	// Add only <n> packets at a time in order to not overload the GUI responsivness
	
	NSMutableArray *array = [NSMutableArray array];
	[mPacketBufferLock lock];
	NSRange r = NSMakeRange(0, MIN([mPacketBuffer count], 100));
	[array addObjectsFromArray:[mPacketBuffer subarrayWithRange:r]];
	[mPacketBuffer removeObjectsInRange:r];
	[mPacketBufferLock unlock];

	// Ignore the GUI if the sniffer is not running (to prevent packets from coming in
	// if the button is "Stop" - because may be all packets haven't been received by this thread)
	if(!mRunning) {
		[self updateStats];
		return;		
	}

	if([array count] == 0) {
		[self updateStats];
		return;
	}
	
	NSArray *selectedObjects = [mResultController selectedObjects];
	
	NSEnumerator *enumerator = [array objectEnumerator];
	id packet;
	int index = [[mResultController content] count];
	while(packet = [enumerator nextObject]) {
		[packet setAbsoluteIndex:++index];
	}

	[mResultController addObjects:array];
	[mResultController rearrangeObjects];
	[mResultController setSelectedObjects:selectedObjects];

	[self updateStats];
	[self updateInfo];
	[self updateChangeCount:NSChangeDone];	
}

// Sniffer delegate

- (void)snifferPacketArrived:(id)packet
{
	[mPacketBufferLock lock];
	[mPacketBuffer addObject:packet];
	[mPacketBufferLock unlock];
}

// NSTableView delegate

- (void)tableViewSelectionDidChange:(NSNotification *)aNotification
{
	[self applyCurrentSelection];
}

@end
