/*
 [The "BSD licence"]
 Copyright (c) 2005 Jean Bovet, http://www.cs.usfca.edu/~jbovet/
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
														   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
														   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "EthernalTool.h"
#include "MoreUNIX.h"
#include "MoreSecurity.h"
#include "MoreCFQ.h"

#include "Constants.h"
#include "BPFLib.h"

#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/uio.h>

#include <net/if.h>
#include <netinet/if_ether.h>
#include <net/bpf.h>

//char		g_en_name[8] = "en1";

static OSStatus GetBPFSocketCommandProc(AuthorizationRef auth, CFDictionaryRef request, CFDictionaryRef *result)
{
    assert(auth != NULL);
    assert(request != NULL);
    assert( result != NULL);
    assert(*result == NULL);
    assert(geteuid() == getuid());
    
	// Get the command string
    CFStringRef command = (CFStringRef) CFDictionaryGetValue(request, kCommandNameKey);
    if(command == NULL || CFGetTypeID(command) != CFStringGetTypeID()) {
		fprintf(stderr, "Command is not of type CFString\n");
        return paramErr;
    }
	
	// Get the slot name (i.e. 'en0' or 'en1' etc)
	CFStringRef slot = (CFStringRef) CFDictionaryGetValue(request, kSlotNameKey);
    if(slot == NULL || CFGetTypeID(slot) != CFStringGetTypeID()) {
		fprintf(stderr, "Slot name is not of type CFString\n");
        return paramErr;
    }	
	
	// Ask for the authorization rights
	static const char *kRightName = "edu.usfca.jean.bovet.EthernalTool";
	static const AuthorizationFlags kAuthFlags = kAuthorizationFlagDefaults
												| kAuthorizationFlagInteractionAllowed
												| kAuthorizationFlagExtendRights
												;
	AuthorizationItem   right  = { kRightName, 0, NULL, 0 };
	AuthorizationRights rights = { 1, &right };
			
	int err = AuthorizationCopyRights(auth, &rights, kAuthorizationEmptyEnvironment, kAuthFlags, NULL);
	if(err != noErr) {
		fprintf(stderr, "EthernalTool: AuthorizationCopyRights() error %d\n", err);
        return err;
	}
	
	// Create the array that will be returned to the parent process
	CFMutableArrayRef	descArray = NULL;
	err = CFQArrayCreateMutable(&descArray);
	if(err != noErr) {
		fprintf(stderr, "EthernalTool: CFQArrayCreateMutable() error %d\n", err);
        return err;
	}
	
	// Make sure we are in privileged mode
	MoreSecSetPrivilegedEUID();

	// Create a BPF socket
	int fd = bpf_new();
	if(fd < 0) {
		fprintf(stderr, "EthernalTool: bpf_new() error %d\n", errno);
        return errno;
	}
	
	err = bpf_set_immediate(fd, 1);
	if(err != noErr) {
		fprintf(stderr, "EthernalTool: bpf_set_immediate() error %d\n", err);
		bpf_dispose(fd);
        return err;
	}
	
	CFIndex slotSize = CFStringGetLength(slot)+1;
	char slotBuffer[slotSize];
	if(!CFStringGetCString(slot, slotBuffer, slotSize, kCFStringEncodingASCII)) {
		fprintf(stderr, "EthernalTool: CFStringGetCString() failed\n");
		bpf_dispose(fd);
        return err;		
	}

	err = bpf_setif(fd, slotBuffer);
	if(err != noErr) {
		fprintf(stderr, "EthernalTool: bpf_setif() error %d\n", err);
		bpf_dispose(fd);
        return err;
	}

	err = bpf_set_promiscuous(fd);
	if(err != noErr) {
		fprintf(stderr, "EthernalTool: bpf_set_promiscuous() error %d\n", err);
		bpf_dispose(fd);
        return err;
	}
	
	// Done making priviledged call
	MoreSecTemporarilySetNonPrivilegedEUID();

	// Create the number to be returned to the parent process	
	CFNumberRef descNum = CFNumberCreate(NULL, kCFNumberIntType, &fd);
	if(descNum == NULL) {
		fprintf(stderr, "EthernalTool: CFNumberCreate() failed\n");
		return -1;
	}

	// Add it to the descriptors array
	CFArrayAppendValue(descArray, descNum);
	CFQRelease(descNum);	
	
	CFStringRef key = kMoreSecFileDescriptorsKey;
	*result = CFDictionaryCreate(NULL, (const void **) &key, (const void **) &descArray, 1, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
	err = CFQError(*result);
	if(err != noErr) {
		fprintf(stderr, "EthernalTool: CFDictionaryCreate() failed\n");
		MoreSecCloseDescriptorArray(descArray);
		CFQRelease(descArray);
		return err;
	}
	CFQRelease(descArray);
	
    return noErr;
}

int main(int argc, const char *argv[])
{
    AuthorizationRef auth = MoreSecHelperToolCopyAuthRef();

    int err = MoreSecDestroyInheritedEnvironment(kMoreSecKeepStandardFilesMask, argv);
	if(err != 0)
		return err;

    // Mask SIGPIPE, otherwise stuff won't work properly.
    err = MoreUNIXIgnoreSIGPIPE();
	if(err != 0)
		return err;

	// Call the MoreSecurity helper routine.
	err = MoreSecHelperToolMain(STDIN_FILENO, STDOUT_FILENO, auth, GetBPFSocketCommandProc, argc, argv);
	if(err != 0)
		return err;
	
    // Map the error code to a tool result.
    return MoreSecErrorToHelperToolResult(err);
}
