//
//  EthernetPacket.m
//  Ethernal
/*
 [The "BSD licence"]
 Copyright (c) 2005 Jean Bovet, http://www.cs.usfca.edu/~jbovet/
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
														   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
														   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "EthernetPacket.h"
#import "Preferences.h"
#import "Constants.h"

@implementation EthernetPacket

+ (EthernetPacket*)packetWithData:(NSData*)data
{
	EthernetPacket *packet = [[EthernetPacket alloc] init];
	[packet setPacket:data];
	return [packet autorelease];
}

+ (EthernetPacket*)packetWithXMLData:(NSData*)data
{
	EthernetPacket *packet = [[EthernetPacket alloc] init];
	[packet setXMLData:data];
	return [packet autorelease];	
}

- (id)init
{
	if(self = [super init]) {
		mPacket = NULL;
		mTime = NULL;
	}
	return self;
}

- (void)dealloc
{
	[mTime release];
	[mPacket release];
	[super dealloc];
}

- (void)setXMLData:(id)data
{
	[mPacket autorelease];
	mPacket = [[data objectForKey:@"packet"] retain];
	[mTime autorelease];
	mTime = [[data objectForKey:@"time"] retain];
}

- (id)xmlData
{
	NSMutableDictionary *dic = [NSMutableDictionary dictionary];
	[dic setObject:mPacket forKey:@"packet"];
	[dic setObject:mTime forKey:@"time"];
	return dic;
}

- (void)setPacket:(NSData*)data
{
	[mPacket autorelease];
	mPacket = [data retain];
}

- (NSString*)getMACStringAtOffset:(int)offset
{
	NSMutableString *mac = [NSMutableString string];
	
	int i;
	for(i = 0; i<6; i++) {
		unsigned char c;
		if(offset+i<[mPacket length])
			[mPacket getBytes:&c range:NSMakeRange(offset+i, 1)];
		[mac appendFormat:@"%02X", c];
		if(i < 5)
			[mac appendString:@":"];
	}
	
	return mac;	
}

- (NSString*)getIPStringAtOffset:(unsigned)offset
{
	if(offset == 0)
		return @"*";
	
	NSMutableString *ip = [NSMutableString string];
	
	int i;
	for(i = 0; i<4; i++) {
		unsigned char c;
		if(offset+i<[mPacket length])
			[mPacket getBytes:&c range:NSMakeRange(offset+i, 1)];
		[ip appendFormat:@"%d", c];
		if(i < 3)
			[ip appendString:@"."];
	}
	
	return ip;	
}

- (void)setAbsoluteIndex:(int)index
{
	mAbsoluteIndex = index;
}

- (int)absoluteIndex
{
	return mAbsoluteIndex;
}

- (void)setTime:(NSString*)time
{
	[mTime autorelease];
	mTime = [time retain];
}

- (NSString*)time
{
	return mTime;
}

- (NSString*)source
{
	return [self getMACStringAtOffset:6];
}

- (NSString*)destination
{
	return [self getMACStringAtOffset:0];
}

- (NSString*)hostNameForIPAddressAtOffset:(unsigned)offset
{
	if(offset<[mPacket length]) {
		unsigned char *bytes = (unsigned char*)[mPacket bytes];
		return [Preferences hostNameForIPAddress:bytes+offset];
	} else
		return @"*";
}

- (NSString*)sourceIP
{
	if([Preferences representIPAddressByValue])
		return [self getIPStringAtOffset:[Preferences sourceIPOffsetForType:[self type]]];
	else
		return [self hostNameForIPAddressAtOffset:[Preferences sourceIPOffsetForType:[self type]]];
}

- (NSString*)destinationIP
{
	if([Preferences representIPAddressByValue])
		return [self getIPStringAtOffset:[Preferences destinationIPOffsetForType:[self type]]];
	else
		return [self hostNameForIPAddressAtOffset:[Preferences destinationIPOffsetForType:[self type]]];
}

- (unsigned)type
{
	unsigned short type = 0;
	if(13<[mPacket length])
		[mPacket getBytes:&type range:NSMakeRange(12, 2)];
	return type;
}

- (NSString*)protocol
{
	if([Preferences representProtocolByValue])
		return [NSString stringWithFormat:@"%04X", [self type]];
	else
		return [Preferences nameOfProtocolType:[self type]];
}

- (int)size
{
	return [mPacket length];
}

- (int)numberOfLineForWidth:(int)width
{
	int line = [mPacket length]/width;
	if([mPacket length] % width != 0)
		line++;
	return line;
}

- (NSString*)addrWidth:(int)width
{
	NSMutableString *s = [NSMutableString string];
	
	int line = [self numberOfLineForWidth:width];
	
	int i;
	for(i=0; i<line; i++) {
		[s appendFormat:@"%04X:\n", i*width];
	}
	
	return s;
}

- (NSString*)asciiWidth:(int)width
{
	NSMutableString *s = [NSMutableString string];
	int i;
	for(i = 0; i<[mPacket length]; i++) {
		unsigned char byte = 0;
		[mPacket getBytes:&byte range:NSMakeRange(i, 1)];
		if(byte < 0x20 || byte > 0x7E)
			byte = '.';
		[s appendFormat:@"%c", byte];
		if((i+1) % width == 0)
			[s appendString:@"\n"];
	}
	return s;
}

- (NSString*)hexaWidth:(int)width
{
	NSMutableString *s = [NSMutableString string];
	int i;
	for(i = 0; i<[mPacket length]; i++) {
		unsigned char byte = 0;
		[mPacket getBytes:&byte range:NSMakeRange(i, 1)];
		[s appendFormat:@"%02X ", byte];
		if((i+1) % width == 0)
			[s appendString:@"\n"];
	}
	return s;
}

- (NSString*)rawHexaContent
{
	NSMutableString *s = [NSMutableString string];
	int i;
	for(i = 0; i<[mPacket length]; i++) {
		unsigned char byte = 0;
		[mPacket getBytes:&byte range:NSMakeRange(i, 1)];
		[s appendFormat:@"%02X", byte];
	}
	return s;	
}

- (NSString*)rawAsciiContentWithoutControlCharacters
{
	NSMutableString *s = [NSMutableString string];
	int i;
	for(i = 0; i<[mPacket length]; i++) {
		unsigned char byte = 0;
		[mPacket getBytes:&byte range:NSMakeRange(i, 1)];
		if(byte < 0x20 || byte > 0x7E)
			byte = '.';
		[s appendFormat:@"%c", byte];
	}
	return s;
}

- (NSString*)rawAsciiContent
{
	return [[[NSString alloc] initWithData:mPacket encoding:NSASCIIStringEncoding] autorelease];
}

@end
